@@ -5,25 +5,30 @@ import uuid
 import os
 import json
 
-from typing import Any, Dict, List, Optional, Tuple
+from typing import Any, Callable, Dict, List, Optional, Tuple
 from contextlib import contextmanager
 
 from inowfaasutils.comm.pubsub import PubSubClient
 
 from ..misc.dataclass_helper import asdict
+from ..callback.base import BaseCallback
+from ..callback import get_callback, get_err_callback
 from ..misc.enum import FaasOpState
 from ..misc.model import Request
+from ..misc.singleton import Singleton
 from .model import FaasError, FaasJob, FaasJobTrigger
 from ..storage.firestore import FireStoreClient
 
 from google.cloud.firestore import (
+    DocumentSnapshot,
     DocumentReference,
+    CollectionReference,
 )
 
 _INIT_JOBS_TOTAL = 1
 
 
-class FaasJobManager:
+class FaasJobManager(metaclass=Singleton):
     """Job completion metadata manager. It is a singleton class used for storing information of nested
     executed FaaS operations, and help to implement architectures like Fan-in and Fan-out.
 
@@ -41,28 +46,30 @@ class FaasJobManager:
     Once all related jobs are finished, end_date is updated with epoch representation (in seconds)
     """
 
-    def __init__(self):
-        self.job_collection_name: str = ""
-        """job collection name in firestore"""
-        self.remaining_job_collection_name: str = ""
-        """remaining job metadata collection name in firestore"""
-        self.failed_job_collection_name: str = ""
-        """failed job metadata collection name in firestore"""
+    firestore: FireStoreClient
+    job_collection_name: str
+    """job collection name in firestore"""
+    remaining_job_collection_name: str
+    """remaining job metadata collection name in firestore"""
+    failed_job_collection_name: str
+    """failed job metadata collection name in firestore"""
+    job_id: str
+    op_id: str
+    has_parent: bool
+    faas_trigger_queue: List[FaasJobTrigger]
+    job_parent_idx_list: Optional[List[int]]
 
+    def __init__(self):
         self._init_envs()
-        self.firestore: FireStoreClient = FireStoreClient(self.GC_PROJECT_ID)
+        self.firestore = FireStoreClient(self.GC_PROJECT_ID)
         self.new_jobs_cnt = 0
         self.diff_increment = 0
-        self.job_id: Optional[str] = None
-        self.op_id: Optional[str] = None
-        self.faas_trigger_queue: List[FaasJobTrigger] = []
-        self.job_parent_idx_list: List[int] = []
-        self.job_done_collection: Optional[str] = None
-        self.user_id = None
-        self.username = None
-        self.job_result = None
+        self.job_id = None
+        self.op_id = None
         self.process_id = None
-        self.parent_job_id = None
+        self.has_parent = False
+        self.faas_trigger_queue = []
+        self.job_parent_idx_list = []
 
     def _init_envs(self):
         """Init all environment variables needed to initialize this class
@@ -98,6 +105,32 @@ class FaasJobManager:
         else:
             raise KeyError("GC_PROJECT_ID not found in environment variables")
 
+    def _end_job(self, req: Request, job_data: FaasJob, last_job: bool = False):
+        """Ends job by adding end_date, and runs callback if any is defined
+
+        Args:
+            req (Request): base Faas Job request
+            job_data (FaasJob): Job metadata
+            last_job (bool, optional): True if is the last job executed in
+            the execution chain. Defaults to False.
+        """
+
+        __end_job: Callable[[DocumentSnapshot], None]
+        if req.callback_type and last_job:
+            callback = get_callback(req)
+
+            @callback.add_callback
+            def __end_job(job_data: FaasJob):
+                job_data.end_date = self._epoch_now()
+
+        else:
+
+            def __end_job(job_data: FaasJob):
+                job_data.end_date = self._epoch_now()
+                return job_data
+
+        return __end_job(job_data)
+
     def _generate_faas_error(self, job_name: str, err: Exception) -> FaasError:
         """Generate FaaS error metadata representation
 
@@ -119,6 +152,19 @@ class FaasJobManager:
             exception_line=err.__traceback__.tb_lineno,
         )
 
+    def _send_err_callback(
+        self, err_data: FaasError, err_callback: Optional[BaseCallback]
+    ):
+        """Send error to a callback function
+
+        Args:
+            err_data (FaasError): FaaS error metadata representation
+            err_callback (Optional[BaseCallback]): Callback function to be run if a
+            job fails
+        """
+
+        err_callback.callback_function(FaasError.Schema().dump(err_data))
+
     def _trigger_faas_queue(self):
         """Enqueue all FaaS calls in the corresponding topics"""
 
@@ -141,6 +187,7 @@ class FaasJobManager:
         req: Request,
         curr_state: FaasOpState,
         err: Exception = None,
+        err_callback: Optional[BaseCallback] = None,
     ):
         """Action after job has been executed
 
@@ -149,6 +196,8 @@ class FaasJobManager:
             req (Request): base Faas Job request
             curr_state (FaasOpState): current state of job execution
             err (Exception, optional): Exception raised on execution. Defaults to None.
+            err_callback (Optional[BaseCallback], optional): Callback function to be run if a
+            job fails. Defaults to None.
         """
 
         state = FaasOpState.SCCS if curr_state != FaasOpState.ERR else FaasOpState.ERR
@@ -164,12 +213,11 @@ class FaasJobManager:
         )
 
         job_data: FaasJob
-        upsert_job_id: str
-        if len(self.job_parent_idx_list) > 0:
+        if self.has_parent:
             parent_doc_ref = self.firestore.get_document_ref(
                 self.job_collection_name,
                 self.job_id,
-                shard_collection_name=self.job_collection_name,
+                collection_name=self.job_collection_name,
                 shard_idx_list=self.job_parent_idx_list[:-1],
             )
             job_data = FaasJob.Schema().load(
@@ -178,29 +226,31 @@ class FaasJobManager:
                     self.job_id,
                     "total_jobs",
                     self.new_jobs_cnt,
-                    shard_collection_name=self.job_collection_name,
+                    collection_name=self.job_collection_name,
                     shard_idx_list=self.job_parent_idx_list,
                 ).to_dict()
             )
-            upsert_job_id = str(self.job_parent_idx_list[-1]) 
         else:
             parent_doc_ref = None
             root_doc.total_jobs += self.new_jobs_cnt
             job_data = root_doc
-            upsert_job_id = self.job_id
 
         job_data.state = state
-        job_data.result = self.job_result
 
         if state == FaasOpState.ERR:
             err_data = self._generate_faas_error(job_name, err)
             self._insert_failed(self.job_id, self.op_id, err_data)
-            self.finish_job(req, root_doc, FaasOpState.ERR)
+            if root_doc.ended is None:
+                root_doc.ended = True
+                self._upsert_job(data=root_doc)
+                if err_callback is not None:
+                    self._send_err_callback(err_data, err_callback)
         elif len(self.faas_trigger_queue) > 0:
             self._trigger_faas_queue()
 
         self._delete_remaining(self.job_id, self.op_id)
 
+        last_job = False
         if (
             not root_doc.ended
             and not self.firestore.has_documents(
@@ -210,17 +260,20 @@ class FaasJobManager:
                 self.failed_job_collection_name, root_doc_ref
             )
         ):
-            self.finish_job(req, root_doc, FaasOpState.SCCS)
+            root_doc.ended = True
+            last_job = True
+            if parent_doc_ref is not None:
+                self._upsert_job(data=root_doc)
 
-        job_data.end_date = self._epoch_now()
-        self._upsert_job(data=job_data, job_id=upsert_job_id, parent_document=parent_doc_ref, result=self.job_result)
+        self._end_job(req, job_data, last_job)
+
+        self._upsert_job(data=job_data, parent_document=parent_doc_ref)
 
     def _get_or_default_job(
         self,
         data: FaasJob = None,
         name: str = None,
         args: Any = None,
-        result: Any = None,
         op_id: str = None,
     ) -> FaasJob:
         """Get FaasJob object given its data. if no data is provideed, then a
@@ -230,7 +283,6 @@ class FaasJobManager:
             data (FaasJob, optional): FaasJob data. Defaults to None.
             name (str, optional): name to set on default FaasJob. Defaults to None.
             args (Any, optional): faas job request input. Defaults to None.
-            result (Any, optional): faas job request result. Defaults to None.
             op_id (str, optional): operation id. Used to check remaining and failed jobs.
             Defaults to None.
         Returns:
@@ -248,9 +300,6 @@ class FaasJobManager:
                 ended=None,
                 total_jobs=_INIT_JOBS_TOTAL,
                 args=args,
-                result=result,
-                process_id=self.process_id,
-                parent_job_id=self.parent_job_id,
             )
             if data is None
             else data
@@ -262,8 +311,6 @@ class FaasJobManager:
         data: FaasJob = None,
         name: str = None,
         args: Any = None,
-        result: Any = None,
-        job_id: str = None,
         parent_document: Optional[DocumentReference] = None,
     ) -> Tuple[str, FaasJob]:
         """Upsert faas job execution metadata into FireStore `job` collection
@@ -273,24 +320,23 @@ class FaasJobManager:
             Defaults to None.
             name (str, optional): faas job first job name. Defaults to None.
             args (Any, optional): faas job request input. Defaults to None.
-            result (Any, optional): faas job request result. Defaults to None.
             parent_document (Optional[DocumentReference]): parent document for nested collections.
             Defaults to None.
         Returns:
             Tuple[str, FaasJob]: id and faas job execution metadata
         """
 
-        job = self._get_or_default_job(data, name, args, result)
+        job = self._get_or_default_job(data, name, args)
         job_dict = asdict(job)
         self.firestore.upsert(
             self.job_collection_name,
-            job_id,
+            self.job_id,
             job_dict,
             parent_document=parent_document,
         )
-        return (job_id, job)
+        return (self.job_id, job)
 
-    def _insert_remaining(self, job_id: str, op_id: str) -> Tuple[str, dict]:
+    def _insert_remaining(self, job_id: str, op_id: str) -> Tuple[str, FaasJob]:
         """Insert faas job remaining execution metadata into FireStore `job.remaining` collection
 
         Args:
@@ -304,12 +350,6 @@ class FaasJobManager:
         self.firestore.upsert(self.remaining_job_collection_name, op_id, {}, doc_ref)
         return (self.op_id, {})
 
-    def _insert_job_done(self, job_done_collection: str, job_args: dict, job_result: dict, status: FaasOpState) -> Tuple[str, dict]:
-        data = {"args": job_args, "result": job_result, "status": status.value}
-        self.firestore.upsert(job_done_collection, self.job_id, data, None)
-        return (self.op_id, data)
-
-
     def _delete_remaining(self, job_id: str, op_id: Optional[str]) -> bool:
         """Deletes metadata of remaining job to be executed
 
@@ -326,7 +366,7 @@ class FaasJobManager:
             remaining_ref: DocumentReference = doc_ref.collection(
                 self.remaining_job_collection_name
             ).document(op_id)
-            self.firestore.delete_document(remaining_ref)
+            return self.firestore.delete_document(remaining_ref)
         return True
 
     def _insert_failed(
@@ -376,43 +416,26 @@ class FaasJobManager:
         """
 
         _job_state = FaasOpState.CRTD
+        err_callback = None
         _err_exception = None
 
         try:
             if req.job_id is None:
                 self.job_id = (str)(uuid.uuid4())
                 self.op_id = None
-                self.user_id = req.user_id
-                self.username = req.username
-                # process id and parent job id must be init
-                # before upsert
-                self.process_id = req.process_id
-                self.parent_job_id = req.parent_job_id
-                self._upsert_job(name=job_name, job_id=self.job_id, args=req)
+                self._upsert_job(name=job_name, args=req)
                 self.job_parent_idx_list = []
-                self.job_done_collection = req.job_done_collection
             else:
                 self.job_id = req.job_id
                 self.op_id = req.op_id
-                self.user_id = req.user_id
-                self.username = req.username
-                self.process_id = req.process_id
-                self.parent_job_id = req.parent_job_id
                 self.job_parent_idx_list = req.job_child_idx_list
-                self.job_done_collection = req.job_done_collection
             yield self
         except Exception as ex:
             _err_exception = ex
             _job_state = FaasOpState.ERR
+            err_callback = get_err_callback(req)
         finally:
-            self._job_close(job_name, req, _job_state, _err_exception)
-
-    def finish_job(self, req: Request, root_job: FaasJob, status: FaasOpState):
-        """Save a job done information in a firestore collection
-        """
-        req.job_id = self.job_id
-        if req.job_done_collection is not None:
-            self._insert_job_done(req.job_done_collection, root_job.args, root_job.result, status)
+            self._job_close(job_name, req, _job_state, _err_exception, err_callback)
 
     def add_job(self, trigger: FaasJobTrigger):
         """Increments total jobs counter on FaasJob metadata, add
@@ -425,15 +448,9 @@ class FaasJobManager:
 
         new_idx_list = self.job_parent_idx_list.copy()
         new_idx_list.append(self.new_jobs_cnt)
-        trigger.message["process_id"] = self.process_id
-        trigger.message["parent_job_id"] = self.parent_job_id
         trigger.message["job_id"] = self.job_id
         trigger.message["op_id"] = (str)(uuid.uuid4())
         trigger.message["job_child_idx_list"] = new_idx_list
-        trigger.message["user_id"] = self.user_id
-        trigger.message["username"] = self.username
-        if self.job_done_collection is not None:
-            trigger.message["job_done_collection"] = self.job_done_collection
         self.new_jobs_cnt += 1
         self._insert_remaining(self.job_id, trigger.message["op_id"])
         job_ref = self.firestore.get_document_ref(
@@ -442,7 +459,7 @@ class FaasJobManager:
             shard_collection_name=self.job_collection_name,
             shard_idx_list=self.job_parent_idx_list,
         )
-        trigger._collection = job_ref.collection(
+        trigger._collection: CollectionReference = job_ref.collection(
             self.job_collection_name
         )
         trigger._job = self._get_or_default_job(
@@ -450,11 +467,3 @@ class FaasJobManager:
         )
         trigger._job_id = str(new_idx_list[-1])
         self.faas_trigger_queue.append(trigger)
-        
-    def set_result(self, result: Any):
-        """Set a result of FaaS Job
-
-        Args:
-            result (Any): result of FaaS Job
-        """
-        self.job_result = result
\ No newline at end of file
