@@ -1,18 +1,23 @@
 from marshmallow_dataclass import dataclass
 from typing import List, Optional
 
+from .enum import CallbackType
+
 
 @dataclass
 class Request:
     """Request base data class"""
-    user_id: Optional[int]
-    username: Optional[str]
-    process_id: Optional[str]
-    parent_job_id: Optional[str]
+
+    callback_type: Optional[CallbackType]
+    callback_location: Optional[str]
+    callback_token: Optional[str]
+    err_callback_type: Optional[CallbackType]
+    err_callback_location: Optional[str]
+    err_callback_token: Optional[str]
     job_id: Optional[str]
     op_id: Optional[str]
     job_child_idx_list: Optional[List[int]]
-    job_done_collection: Optional[str]
+
 
 @dataclass
 class Response:
@@ -30,3 +35,9 @@ class ResponseError(Response):
     error: str
     code: int
 
+
+@dataclass
+class CallbackCreation:
+    type: Optional[CallbackType]
+    location: Optional[str]
+    token: Optional[str]
