@@ -50,8 +50,7 @@ class FireStoreClient(metaclass=Singleton):
             )
             doc_id = id if id is not None else "self"
             root_ref.collection(collection_name).document(doc_id).set(
-                data.to_dict() if isinstance(data, DocumentSnapshot) else data,
-                merge=True
+                data.to_dict() if isinstance(data, DocumentSnapshot) else data
             )
             return True
         except Exception as ex:
@@ -115,14 +114,17 @@ class FireStoreClient(metaclass=Singleton):
         except Exception:
             return None
 
-    def delete_document(self, doc_ref: DocumentReference):
+    def delete_document(self, doc_ref: DocumentReference) -> bool:
         """Deletes document
 
         Args:
             doc_ref (DocumentReference): _description_
+
+        Returns:
+            bool: _description_
         """
 
-        doc_ref.delete()
+        return doc_ref.delete().seconds > 0
 
     def has_documents(
         self, collection_name: str, parent_document: Optional[DocumentReference] = None
